/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                         Common.h                        **/
/**                                                         **/
/** This file contains screen refresh drivers which are     **/
/** common for both X11 and VGA implementations. It is      **/
/** included either from Unix.c or MSDOS.c.                 **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994,1995                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

int ExitFlag=0;
int IFreq=50;

byte *XBuf;
byte XPal[16],XPal0,Pal[16][3];


int InitScrF(void)
{ 
  if(Verbose>1) printf("Screen %d initialized\n",ScrMode);
  return(1);
}


void Sound(void)
{
  byte J;
  if(Verbose>1)
    for(J=0;J<3;J++)
      if(Volume[J])
        printf("CHANNEL %c: F=%d V=%d\n",J+'A',Freq[J],Volume[J]);
}


void Sprites(void)
{
  register byte N,M,I,J,X,Y,C,H;
  register byte *P,*T,*S;
  register int L;

  for(N=0,S=SprTab;(N<32)&&(S[0]!=208);N++,S+=4);
  H=ScanLines212? 212:192;
  M=SolidColor0;
  
  if(Sprites16x16)
    for(S-=4;N;N--,S-=4)
    {
      C=S[3]&0x0F;L=S[3]&0x80? S[1]-32:S[1];
      if((L<=256-16)&&(L>=0)&&(C||M))
      {
        P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2+L;
        T=SprGen+((long)(S[2]&0xFC)<<3);
        C=XPal[C];Y=S[0]+1;

        if(Y<H) { P+=WIDTH*Y;Y=(Y>H-16)? H-Y:16; } 
        else    { T+=256-Y;Y-=(Y>240)? 240:Y; }

        for(;Y;Y--,T++,P+=WIDTH-8)
          for(X=0,J=*T,I=*(T+16);X<8;X++,J<<=1,I<<=1,P++)
          {
            if(J&0x80) *P=C;
            if(I&0x80) *(P+8)=C;
          }          
      }
    }
  else
    for(S-=4;N;N--,S-=4)
    {
      C=S[3]&0x0F;L=S[3]&0x80? S[1]-32:S[1];
      if((L<=256-8)&&(L>=0)&&(C||M))
      {
        P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2+L;
        T=SprGen+((long)S[2]<<3);C=XPal[C];Y=S[0]+1;
     
        if(Y<H) { P+=WIDTH*Y;Y=(Y>H-8)? H-Y:8; }
        else    { T+=256-Y;Y-=(Y>248)? 248:Y; }

        for(;Y;Y--,T++,P+=WIDTH-8)
          for(X=0,J=*T;X<8;X++,J<<=1,P++)
            if(J&0x80) *P=C;
      }
    }
}


void ColorSprites(void)
{
  register byte N,M,I,J,X,Y,C,H;
  register byte *P,*T,*AT,*PT,*CT;
  register int L;

  for(N=0,AT=SprTab;(N<32)&&(AT[0]!=216);N++,AT+=4);
  CT=SprTab-0x0200+((long)N<<4)-16;AT-=4;
  H=ScanLines212? 212:192;
  M=SolidColor0;

  if(Sprites16x16)
    for(;N;N--,AT-=4,CT-=16)
    {
      P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2;
      PT=SprGen+((long)(AT[2]&0xFC)<<3);T=CT;

      Y=AT[0]+1-VDP[23];      
      if(Y<H) { P+=WIDTH*Y;Y=(Y>H-16)? H-Y:16; }
      else    { Y=256-Y;PT+=Y;T+=Y;Y=(Y<16)? 16-Y:0; }      

      for(;Y;Y--,T++,PT++,P+=WIDTH)
      {
        L=*T&0x80? AT[1]-32:AT[1];C=*T&0x0F;
        if((L<=256-16)&&(L>=0)&&(C||M))
        {
          P+=L;J=*PT;I=*(PT+16);C=XPal[C];
          for(X=8;X;X--,J<<=1,I<<=1,P++)
          {
            if(J&0x80) *P=C;
            if(I&0x80) *(P+8)=C;
          }
          P-=8+L;
        }
      }
    }
  else
    for(;N;N--,AT-=4,CT-=16)
    {
      P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2;
      PT=SprGen+((int)AT[2]<<3);T=CT;

      Y=AT[0]+1-VDP[23];
      if(Y<H) { P+=WIDTH*Y;Y=(Y>H-8)? H-Y:8; }
      else    { Y=256-Y;PT+=Y;T+=Y;Y=(Y<8)? 8-Y:0; }

      for(;Y;Y--,T++,PT++,P+=WIDTH)
      {
        L=*T&0x80? AT[1]-32:AT[1];C=*T&0x0F;
        if((L<=256-8)&&(L>=0)&&(C||M))
        {
          P+=L;J=*PT;C=XPal[C];
          for(X=8;X;X--,J<<=1,P++)
            if(J&0x80) *P=C;
          P-=8+L;
        }
      }
    }
}


void RefreshScrF(void)
{
  if(Verbose>1)
    printf
    (
      "ScrMODE %d: ChrTab=%X ChrGen=%X ColTab=%X SprTab=%X SprGen=%X\n",
      ScrMode,ChrTab-VRAM,ChrGen-VRAM,ColTab-VRAM,SprTab-VRAM,SprGen-VRAM
    );
  memset(XBuf,XPal[BGColor],HEIGHT*WIDTH);
}


void RefreshScr0(void)
{
  register byte X,Y,J,K,FC,BC;
  register byte *P,*S,*T;

  BC=XPal[BGColor];FC=XPal[FGColor];

  if(ScreenON)
  {
    P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-240)/2;T=ChrTab;

    for(Y=ScanLines212? 26:24;Y;Y--,P+=WIDTH*8-6*40)
      for(X=0;X<40;X++,T++)
      {
        S=ChrGen+((long)*T<<3);
        for(J=0;J<8;J++)
        {
          K=*S++;
          *P++=K&0x80? FC:BC;*P++=K&0x40? FC:BC;
          *P++=K&0x20? FC:BC;*P++=K&0x10? FC:BC;
          *P++=K&0x08? FC:BC;*P  =K&0x04? FC:BC;
          P+=WIDTH-5;
        }
        P+=6-WIDTH*8;
      }
  }
  else memset(XBuf,BC,WIDTH*HEIGHT);

  PutImage;
}


void RefreshScr1(void)
{
  register byte X,Y,K,J,FC,BC;
  register byte *P,*S,*T;

  if(ScreenON)
  {
    P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2;T=ChrTab;

    for(Y=ScanLines212? 26:24;Y;Y--,P+=WIDTH*8-8*32)
      for(X=0;X<32;X++,T++)
      {
        S=ChrGen+((long)*T<<3);
        BC=*(ColTab+(*T>>3));
        FC=XPal[BC>>4];BC=XPal[BC&0x0F];
        for(J=0;J<8;J++)
	{
          K=*S++;
          *P++=K&0x80? FC:BC;*P++=K&0x40? FC:BC;
          *P++=K&0x20? FC:BC;*P++=K&0x10? FC:BC;
          *P++=K&0x08? FC:BC;*P++=K&0x04? FC:BC;
          *P++=K&0x02? FC:BC;*P  =K&0x01? FC:BC;
          P+=WIDTH-7;  
        }
        P+=8-WIDTH*8;
      }
  }
  else memset(XBuf,XPal[BGColor],HEIGHT*WIDTH);

  if(!SpritesOFF) Sprites();

  PutImage;
}


void RefreshScr2(void)
{
  register byte X,Y,J,K,M,N,FC,BC;
  register byte *P,*S,*T,*PGT,*CLT,*C;

  if(ScreenON)
  {
    P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2;
    PGT=ChrGen;CLT=ColTab;T=ChrTab;
    M=ScanLines212? 4:3;

    for(N=0;N<M;N++,PGT+=0x0800,CLT+=0x0800)
      for(Y=(N==3)? 2:8;Y;Y--,P+=WIDTH*8-8*32)
        for(X=0;X<32;X++,T++)
        {
          S=PGT+((long)*T<<3);C=CLT+((long)*T<<3);
          for(J=0;J<8;J++)
          {
            BC=*C++;K=*S++;
            FC=XPal[BC>>4];BC=XPal[BC&0x0F];
            *P++=K&0x80? FC:BC;*P++=K&0x40? FC:BC;
            *P++=K&0x20? FC:BC;*P++=K&0x10? FC:BC;
            *P++=K&0x08? FC:BC;*P++=K&0x04? FC:BC;
            *P++=K&0x02? FC:BC;*P  =K&0x01? FC:BC;
            P+=WIDTH-7;
          }
          P+=8-WIDTH*8;
        }
  }
  else memset(XBuf,XPal[BGColor],HEIGHT*WIDTH);

  if(!SpritesOFF) Sprites();

  PutImage;
}


void RefreshScr3(void)
{
  register byte X,Y,J,M,C1,C2;
  register byte *P,*T,*C;

  if(ScreenON)
  {
    P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2;
    T=ChrTab;C=ColTab;M=ScanLines212? 26:24;

    for(Y=0;Y<M;Y++,P+=WIDTH*8-8*32)
      for(X=0;X<32;X++,T++)
      {
        C=ColTab+((long)*T<<3)+((Y&3)<<1);
        C1=XPal[*C>>4],C2=XPal[*C&15];
        for(J=0;J<4;J++,P+=WIDTH-7)
        { 
          *P++=C1;*P++=C1;*P++=C1;*P++=C1;
          *P++=C2;*P++=C2;*P++=C2;*P=C2;
        }
        C++;C1=XPal[*C>>4],C2=XPal[*C&15];
        for(J=0;J<4;J++,P+=WIDTH-7)
        {
          *P++=C1;*P++=C1;*P++=C1;*P++=C1;
          *P++=C2;*P++=C2;*P++=C2;*P=C2;
        }
        P+=8-8*WIDTH;  
      }
  }
  else memset(XBuf,XPal[BGColor],HEIGHT*WIDTH);

  if(!SpritesOFF) Sprites();

  PutImage;
}


void RefreshScr4(void)
{
  register byte X,Y,J,K,M,N,FC,BC;
  register byte *P,*S,*T,*PGT,*CLT,*C;

  if(ScreenON)
  {
    P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2;
    PGT=ChrGen;CLT=ColTab;T=ChrTab;
    M=ScanLines212? 4:3;

    for(N=0;N<M;N++,PGT+=0x0800,CLT+=0x0800)
      for(Y=(N==3)? 2:8;Y;Y--,P+=WIDTH*8-8*32)
        for(X=0;X<32;X++,T++)
        {
          S=PGT+((long)*T<<3);C=CLT+((long)*T<<3);
          for(J=0;J<8;J++)
          {
            BC=*C++;K=*S++;
            FC=XPal[BC>>4];BC=XPal[BC&0x0F];
            *P++=K&0x80? FC:BC;*P++=K&0x40? FC:BC;
            *P++=K&0x20? FC:BC;*P++=K&0x10? FC:BC;
            *P++=K&0x08? FC:BC;*P++=K&0x04? FC:BC;
            *P++=K&0x02? FC:BC;*P  =K&0x01? FC:BC;
            P+=WIDTH-7;
          }
          P+=8-WIDTH*8;
        }
  }
  else memset(XBuf,XPal[BGColor],HEIGHT*WIDTH);

  if(!SpritesOFF) ColorSprites();

  PutImage;
}


void RefreshScr5(void)
{
  register byte X,Y,J;
  register byte *P,*T;

  if(ScreenON)
  {
    P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2;
    T=ChrTab+((long)VDP[23]<<7);
    Y=ScanLines212? 212:192;

    for(J=VDP[23];Y;Y--,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<128;X++,T++)
      { *P++=XPal[*T>>4];*P++=XPal[*T&0x0F]; }
    }
  }
  else memset(XBuf,XPal[BGColor],HEIGHT*WIDTH);

  if(!SpritesOFF) ColorSprites();

  PutImage;
}


void RefreshScr6(void)
{
  register byte X,Y,J;
  register byte *P,*T;

  if(ScreenON)
  {
    P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2;
    T=ChrTab+((long)VDP[23]<<7);
    Y=ScanLines212? 212:192;

    for(J=VDP[23];Y;Y--,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<128;X++,T++)
      { *P++=XPal[*T>>6];*P++=XPal[(*T>>2)&0x03]; }
    }
  }
  else memset(XBuf,XPal[BGColor],HEIGHT*WIDTH);

  if(!SpritesOFF) ColorSprites();

  PutImage;
}


void RefreshScr7(void)
{
  register byte X,Y,J;
  register byte *P,*T;

  if(ScreenON)
  {
    P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2;
    T=ChrTab+((long)VDP[23]<<8);
    Y=ScanLines212? 212:192;

    for(J=VDP[23];Y;Y--,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<128;X++)
      { *P++=XPal[*T>>4];T++;*P++=XPal[*T>>4];T++; }
    }
  }
  else memset(XBuf,XPal[BGColor],HEIGHT*WIDTH);

  if(!SpritesOFF) ColorSprites();

  PutImage;
}


void RefreshScr8(void)
{
  register byte Y,*P,*T,J;
  register word X;

  if(ScreenON)
  {
    P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-256)/2;
    T=ChrTab+((long)VDP[23]<<8);
    Y=ScanLines212? 212:192;

    for(J=VDP[23];Y;Y--,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<256;X++,T++) *P++=*T;
    }
  }
  else memset(XBuf,XPal[BGColor],HEIGHT*WIDTH);

  if(!SpritesOFF) ColorSprites();

  PutImage;
}


void RefreshTx80(void)
{
  register byte X,Y,J,K,M,FC,BC;
  register byte *P,*S,*C,*T;

  if(ScreenON)
  {
    P=XBuf+WIDTH*(HEIGHT-192)/2+(WIDTH-320)/2;
    C=ColTab;T=ChrTab;M=0x80;

    for(Y=ScanLines212? 26:24;Y;Y--,P+=WIDTH*8-4*80)
      for(X=0;X<80;X++,T++)
      {
        S=ChrGen+((long)*T<<3);
        if(*C&M) { FC=XPal[FGColor];BC=XPal[BGColor]; }
        else     { FC=XPal[XFGColor];BC=XPal[XBGColor]; }
        for(J=0;J<8;J++)
        {
          K=*S++;
          *P++=K&0xC0? FC:BC;
          *P++=K&0x30? FC:BC;
          *P++=K&0x0C? FC:BC;
          *P=BC;P+=WIDTH-3;
        }
        P+=4-WIDTH*8;
        if(!(M>>=1)) { M=0x80;C++; } 
      }
  }
  else memset(XBuf,XPal[BGColor],HEIGHT*WIDTH);

  PutImage;
}
