/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                           MSX.h                         **/
/**                                                         **/
/** This file contains declarations relevant to the drivers **/
/** and MSX emulation itself. See Z80.h for #defines        **/
/** related to Z80 emulation.                               **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994,1995                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

/* #define UNIX  */         /* Compile fMSX for for Unix/X   */
/* #define MSDOS */         /* Compile fMSX for MSDOS/VGA    */
#define MITSHM              /* Use MIT SHM extensions for X  */

#ifndef UNIX
#undef MITSHM
#endif

#define PAGESIZE  0x4000L   /* Size of a RAM page            */
#define NORAM     0xFF      /* Byte to be returned from      */
                            /* non-existing pages and ports  */
#define MAXSCREEN 8         /* Highest screen mode supported */
#define SNDCONST  3579000L  /* Sound constant 3.579MHz/8 */

/***** Following are macros to be used in screen drivers *****/
#define BigSprites    (VDP[1]&0x01)   /* Zoomed sprites      */
#define Sprites16x16  (VDP[1]&0x02)   /* 16x16/8x8 sprites   */
#define SpritesOFF    (VDP[8]&0x02)   /* Don't show sprites  */
#define ScreenON      (VDP[1]&0x40)   /* Show screen         */
#define SolidColor0   (VDP[8]&0x20)   /* Solid/Tran. COLOR 0 */
#define ScanLines212  (VDP[9]&0x80)   /* 212/192 scanlines   */
/*************************************************************/

#include "Z80.h"            /* Z80 emulation declarations    */

/******** Variables used to control emulator behavior ********/
extern byte Verbose;                  /* Debug msgs ON/OFF   */
extern byte MSXVersion;               /* 0=MSX1, 1=MSX2      */
extern byte RAMPages,VRAMPages;       /* Number of RAM pages */
/*************************************************************/

extern byte *VRAM,*RAM;               /* Main and Video RAMs */

extern byte EnWrite[4];               /* 1 if write enabled  */

extern byte KeyMap[16];               /* Keyboard map        */

extern char *CartA;                   /* Cartridge A ROM file*/
extern char *CartB;                   /* Cartridge B ROM file*/

extern char *PrnName;                 /* Printer redir. file */

extern byte UPeriod;  /* Number interrupts per screen update */

extern byte *ChrGen,*ChrTab,*ColTab;  /* VDP tables [screens]*/
extern byte *SprGen,*SprTab,*SprCol;  /* VDP tables [sprites]*/
extern byte Palette[32][3];           /* Palette registers   */
extern byte FGColor,BGColor;          /* Colors              */
extern byte XFGColor,XBGColor;        /* Second set of colors*/
extern byte ScrMode;                  /* Current screen mode */
extern byte VDP[64],VDPStatus[16];    /* VDP registers       */

extern int  Volume[3],Freq[3];        /* Sound parameters    */

/****************************************************************/
/*** These functions are called to initialize and refresh     ***/
/*** various SCREENs.                                         ***/
/************************************** TO BE WRITTEN BY USER ***/
int InitTx80(void);void ColorsTx80(void);void RefreshTx80(void);
int InitScr0(void);void ColorsScr0(void);void RefreshScr0(void);
int InitScr1(void);void ColorsScr1(void);void RefreshScr1(void);
int InitScr2(void);void ColorsScr2(void);void RefreshScr2(void);
int InitScr3(void);void ColorsScr3(void);void RefreshScr3(void);
int InitScr4(void);void ColorsScr4(void);void RefreshScr4(void);
int InitScr5(void);void ColorsScr5(void);void RefreshScr5(void);
int InitScr6(void);void ColorsScr6(void);void RefreshScr6(void);
int InitScr7(void);void ColorsScr7(void);void RefreshScr7(void);
int InitScr8(void);void ColorsScr8(void);void RefreshScr8(void);
int InitScrF(void);void ColorsScrF(void);void RefreshScrF(void);

/****************************************************************/
/*** This function is called to poll keyboard.                ***/
/************************************** TO BE WRITTEN BY USER ***/
void Keyboard(void);

/****************************************************************/
/*** This function is called to generate sound.               ***/
/************************************** TO BE WRITTEN BY USER ***/
void Sound(void);

/****************************************************************/
/*** Allocate memory, load ROM images, initialize mapper, VDP ***/
/*** CPU and start the emulation. This function returns 0 in  ***/
/*** the case of failure.                                     ***/
/****************************************************************/
int StartMSX(void);

/****************************************************************/
/*** Free memory allocated by StartMSX().                     ***/
/****************************************************************/
void TrashMSX(void);

/****************************************************************/
/*** Set sound volumes, refresh screen, check keyboard and    ***/
/*** sprites. Call this function on each interrupt.           ***/
/****************************************************************/
void Interrupt(void);

/****************************************************************/
/*** Switch secondary memory slots. This function is supposed ***/
/*** to be called when value in (FFFFh) changes.              ***/
/****************************************************************/
void SSlot(byte Value);
