/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                           Unix.c                        **/
/**                                                         **/
/** This file contains Unix/X-dependent subroutines and     **/
/** drivers. It includes common drivers from Common.h.      **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994,1995                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include "MSX.h"

#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
XShmSegmentInfo SHMInfo;
int UseSHM=1;
#endif

char *Title="fMSX Unix/X 0.7";

struct itimerval TimerValue;

Screen *screen;
Display *disp;
Window root,win;
Colormap cmap;
XImage *ximage;
XEvent ev;
GC defaultGC,gc;
int screen_num;
unsigned long white,black;

#ifdef MITSHM
#define PutImage \
  if(UseSHM)                                                            \
    XShmPutImage(disp,win,defaultGC,ximage,0,0,0,0,WIDTH,HEIGHT,False); \
  else                                                                  \
    XPutImage(disp,win,defaultGC,ximage,0,0,0,0,WIDTH,HEIGHT);          \
  XFlush(disp)
#else
#define PutImage \
  XPutImage(disp,win,defaultGC,ximage,0,0,0,0,WIDTH,HEIGHT); \
  XFlush(disp);
#endif

#define WIDTH  340
#define HEIGHT 240
#include "Common.h"

void OnBreak(void) { ExitFlag=1; }
void OnTimer(void) { IFlag=1; }

byte Keys[256][2] =
{
/* 00 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {7,0x20},{7,0x08},{0,0x00},{0,0x00},{0,0x00},{7,0x80},{0,0x00},{0,0x00},
/* 10 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{7,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 20 */
  {8,0x01},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x01},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x04},{1,0x04},{2,0x08},{2,0x10},
/* 30 */
  {0,0x01},{0,0x02},{0,0x04},{0,0x08},{0,0x10},{0,0x20},{0,0x40},{0,0x80},
  {1,0x01},{1,0x02},{0,0x00},{1,0x80},{0,0x00},{1,0x08},{0,0x00},{0,0x00},
/* 40 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 50 */
  {0,0x00},{8,0x10},{8,0x20},{8,0x80},{8,0x40},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{1,0x20},{1,0x10},{1,0x40},{0,0x00},{0,0x00},
/* 60 */
  {2,0x02},{2,0x40},{2,0x80},{3,0x01},{3,0x02},{3,0x04},{3,0x08},{3,0x10},
  {3,0x20},{3,0x40},{3,0x80},{4,0x01},{4,0x02},{4,0x04},{4,0x08},{4,0x10},
/* 70 */
  {4,0x20},{4,0x40},{4,0x80},{5,0x01},{5,0x02},{5,0x04},{5,0x08},{5,0x10},
  {5,0x20},{5,0x40},{5,0x80},{0,0x00},{0,0x00},{0,0x00},{6,0x04},{0,0x00},
/* 80 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 90 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* A0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* B0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{6,0x20},{6,0x40},
/* C0 */
  {6,0x80},{7,0x01},{7,0x02},{8,0x02},{8,0x08},{8,0x04},{7,0x40},{7,0x10},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* D0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {8,0x02},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* E0 */
  {0,0x00},{6,0x01},{6,0x08},{6,0x02},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{6,0x10},{6,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* F0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{8,0x08}
};


void TrashX(void)
{
  if(disp&&win)
  {
#ifdef MITSHM
    if(UseSHM)
    {
      XShmDetach(disp,&SHMInfo);
      if(SHMInfo.shmaddr) shmdt(SHMInfo.shmaddr);
      if(SHMInfo.shmid>=0) shmctl(SHMInfo.shmid,IPC_RMID,0);
    }
    else
#endif MITSHM
    if(ximage) XDestroyImage(ximage);
    XAutoRepeatOn(disp);
  }
  if(disp) XCloseDisplay(disp);
}


int InitX(void)
{
  XGCValues values;
  int J;

  if(Verbose)
    printf("Initializing X11 drivers:\n  Opening display...");
  disp=XOpenDisplay(NULL);
  if(!disp) { if(Verbose) printf("FAILED\n");return(0); }

  screen=DefaultScreenOfDisplay(disp);screen_num=DefaultScreen(disp);
  white=XWhitePixel(disp,screen_num);black=XBlackPixel(disp,screen_num);
  root=DefaultRootWindow(disp);
  defaultGC=XDefaultGC(disp,screen_num);
  cmap=DefaultColormap(disp,screen_num);

  if(Verbose) printf("OK\n  Opening window...");
  win=XCreateSimpleWindow(disp,root,0,0,WIDTH,HEIGHT,0,white,black);
  if(!win) { if(Verbose) printf("FAILED\n");return(0); }

  {
    XSizeHints hints;
    XWMHints wmhints;

    hints.flags=USSize;hints.width=WIDTH;hints.height=HEIGHT;
    wmhints.input=True;wmhints.flags=InputHint;
    XSetWMHints(disp,win,&wmhints);
    XSetWMNormalHints(disp,win,&hints);
    XStoreName(disp,win,Title);
  }

  XSelectInput(disp,win,ExposureMask|KeyPressMask|KeyReleaseMask);
  XMapRaised(disp,win);XNextEvent(disp,&ev);

  gc=XCreateGC(disp,win,0,&values);
  XSetForeground(disp,gc,white);
  XSetBackground(disp,gc,black);
  XClearWindow(disp,win);

  XAutoRepeatOff(disp);

#ifdef MITSHM
  if(UseSHM)
  {
    if(Verbose)
      printf("OK\n  Using shared memory:\n    Creating image...");
    ximage=XShmCreateImage
           (
             disp,DefaultVisual(disp,screen_num),8,
             ZPixmap,NULL,&SHMInfo,WIDTH,HEIGHT
           );
    if(!ximage) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n    Getting SHM info...");
    SHMInfo.shmid=shmget
                  (
                    IPC_PRIVATE,ximage->bytes_per_line*ximage->height,
                    IPC_CREAT|0777
                  );
    if(SHMInfo.shmid<0) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n    Allocating SHM...");
    XBuf=(byte *)(ximage->data=SHMInfo.shmaddr=shmat(SHMInfo.shmid,0,0));
    if(!XBuf) { if(Verbose) printf("FAILED\n");return(0); }

    SHMInfo.readOnly=False;
    if(Verbose) printf("OK\n    Attaching SHM...");
    if(!XShmAttach(disp,&SHMInfo))
    { if(Verbose) printf("FAILED\n");return(0); }
  }
  else
#endif
  {
    if(Verbose) printf("OK\n  Allocating RAM...");
    XBuf=(byte *)malloc(sizeof(byte)*HEIGHT*WIDTH);
    if(!XBuf) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n  Creating image...");
    ximage=XCreateImage
           (
             disp,DefaultVisual(disp,screen_num),8,
             ZPixmap,0,XBuf,WIDTH,HEIGHT,8,0
           );
    if(!ximage) { if(Verbose) printf("FAILED\n");return(0); }
  }

  for(XPal0=black,J=0;J<16;J++)
  { memcpy(Pal[J],"\0\0\0",3);XPal[J]=black; }

  if(Verbose) printf("OK\n");

  if(!IntSync)
  {
    if(Verbose) printf("Initializing timer for %dHz...",IFreq);
    TimerValue.it_interval.tv_sec =
    TimerValue.it_value.tv_sec    = 0L;
    TimerValue.it_interval.tv_usec=
    TimerValue.it_value.tv_usec   = 1000000L/IFreq;
    J=setitimer(ITIMER_VIRTUAL,&TimerValue,NULL);
    if(Verbose) printf(J? "FAILED\n":"OK\n");
    if(J) return(0);
    signal(SIGVTALRM,OnTimer);
  }

  signal(SIGHUP,OnBreak);signal(SIGINT,OnBreak);
  signal(SIGQUIT,OnBreak);signal(SIGTERM,OnBreak);

  return(1);
}


void Keyboard(void)
{
  byte J;

  if(XCheckWindowEvent(disp,win,KeyPressMask|KeyReleaseMask,&ev))
  {
    J=XLookupKeysym((XKeyEvent *)&ev,0)&0xFF;
    if(ev.type==KeyPress)
    {
#ifdef DEBUG
      if(J==0xC9) Trace=!Trace;
#endif
      KeyMap[Keys[J][0]]&=~Keys[J][1];
    }
    else KeyMap[Keys[J][0]]|=Keys[J][1];
  }

  if(ExitFlag)
  {
    if(Verbose) printf("Shutting down...\n"); 
    TrashMSX();TrashX();exit(0);
  }
}


void ColorsScrF(void)
{
  XColor Color;
  unsigned long L;
  byte *P,C;
  int J;

  XPal[0]=XPal0;
  for(J=0;J<16;J++)
    if(memcmp(Palette[J],Pal[J],3))
    {
      memcpy(Pal[J],Palette[J],3);L=XPal[J];
      if(L!=black) XFreeColors(disp,cmap,&L,1,0);
      Color.flags=DoRed|DoGreen|DoBlue;
      Color.red=Palette[J][0]*256;
      Color.green=Palette[J][1]*256;
      Color.blue=Palette[J][2]*256;
      if(!memcmp(Pal[J],"\0\0\0",3)) XPal[J]=black;
      else
        if(XAllocColor(disp,cmap,&Color)) XPal[J]=Color.pixel;
        else { XPal[J]=black;memcpy(Pal[J],"\0\0\0",3); }

    }
  XPal0=XPal[0];

  C=XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];
  for(J=0,P=XBuf;J<HEIGHT*WIDTH;J++) *P++=C;
}
